local revision = tonumber(string.sub("$Revision: 68 $", 12, -3))
local rss = _G.RShamanStat
if rss.revision < revision then rss.revision = revision end

--Monitor and record player gear, stat, and buff changes

local playerState = 
{
	GUID = nil,
	role = "UNKNOWN",
	isGoblin = nil,
	level = nil,
	ascendance = nil,
	ascendanceAura = false,
	ascendanceCast = nil,
	ancestralGuidance = nil,
	ancestralGuidanceCast = nil,
	guidanceAura = false,
	tidalWaves = nil,
	healingBonus = 0,
	damageBonus = { [1] = 0.0, [2] = 0.0, [3] = 0.0, [4] = 0.0, [5] = 0.0, [6] = 0.0, [7] = 0.0 },
	masteryBonus = 0.0,
	criticalChance = { [1] = 0.0, [2] = 0.0, [3] = 0.0, [4] = 0.0, [5] = 0.0, [6] = 0.0, [7] = 0.0 },
	criticalMultiplier = rss.CRITICAL_MULTIPLIER,
	hasteRating = 0,
	baseHasteRating = 8930,
	targetHasteRating = 10513,
	spellHaste = 0.0,
	buffedSpellHaste = 0.0,
	spellHasteBuffActive = false,
	ancestralSwiftnessActive = nil,
	hasteBreakpoints = nil,
	aaSetBonusT15 = false,
	spiritLinkGUID = nil,
	lastCast = { },
	lastFlameShock = { combatEvent = "SPELL_DAMAGE" },
	busy = 0,
	busyRing = nil,
	busyCounts = 0,
	busyPointer = 0,
	busyInterval = 1.0/rss.WINDOW_RATE,
	busyElapsed = 0.0,
}
rss.playerState = playerState;

local healthTable = { };
rss.healthTable = healthTable;

function rss.CreateCopyRecord()
	local record = { };
	record.rawHealing = { 0.0 , 0.0 };
	record.incrementalSp = 0.0;
	record.incrementalMastery = 0.0;
	record.incrementalCritical = { 0.0 , 0.0 };
	record.masteryTotal = 0.0;
	record.effectiveCopies = 0.0;
	return record;
end

-- deal with healing copy abilities
-- we assume that every 1% increase in our raw (not effective) heals during the duration will mean a corresponding 1% increase in the effective (not raw) copied heals
function rss.ProcessCopyRecord(record, fight)
	local totalRaw = record.rawHealing[1] + record.rawHealing[2];
	if fight and totalRaw > 0 then
		fight.incrementalSpellpower = fight.incrementalSpellpower + record.effectiveCopies*(record.incrementalSp/totalRaw);
		record.incrementalSp = 0.0;
		fight.incrementalMastery = fight.incrementalMastery + record.effectiveCopies*(record.incrementalMastery/totalRaw);
		record.incrementalMastery = 0.0;
		fight.masteryTotal = fight.masteryTotal + record.masteryTotal * (record.effectiveCopies/totalRaw);
		record.masteryTotal = 0.0;
		fight.criticalTotal = fight.criticalTotal + record.incrementalCritical[1]*(record.rawHealing[1]/totalRaw)*(rss.RATING_TO_CRIT/0.01)
		for i = 1, 2 do
			fight.incrementalCritical = fight.incrementalCritical + record.incrementalCritical[i]*(record.rawHealing[i]/totalRaw);
			record.incrementalCritical[i] = 0.0;
		end
		record.rawHealing = { 0.0 , 0.0 };
		record.effectiveCopies = 0.0;
	end
end

function rss.ResetPlayerState()
	playerState.GUID = UnitGUID("player");
	playerState.isGoblin = select(2, UnitRace("player")) == "Goblin";
	playerState.level = UnitLevel("player");
	playerState.baseHasteRating = GetCombatRating(CR_HASTE_SPELL);
	rss.CheckGear();
	rss.CheckTalents();
	rss.ClearHealthTable();
	playerState.lastCast = { };
	playerState.spiritLinkGUID = nil;
	playerState.ascendance = nil;
	playerState.ascendanceCast = nil;
	playerState.ancestralGuidance = nil;
	playerState.ancestralGuidanceCast = nil;
end

function rss.MonitorState(fight)
	playerState.healingBonus = GetSpellBonusHealing();
	for i = 1, 7 do
		playerState.damageBonus[i] = GetSpellBonusDamage(i);
	end
	playerState.masteryBonus = 0.03 * GetMastery();
	for i = 1, 7 do
		playerState.criticalChance[i] = GetSpellCritChance(i)/100.;
	end
	playerState.spellHaste = UnitSpellHaste("player")/100.0;
	playerState.hasteRating = GetCombatRating(CR_HASTE_SPELL);
	
	local tidalWavesUp = false;
	local ascendanceUp = false;
	local guidanceUp = false;
	playerState.spellHasteBuffActive = false;
	for i = 1, 40 do
		local id = select(11,UnitAura("player", i));
		if id then
			if id == rss.AURA_TIDAL_WAVES then
				tidalWavesUp = true;
			elseif id == rss.AURA_ASCENDANCE then
				ascendanceUp = true;
			elseif id == rss.AURA_ANCESTRAL_GUIDANCE then
				guidanceUp = true;
			elseif tContains(rss.HASTE_BUFFS, id) then
				playerState.spellHasteBuffActive = true;
			end
		else
			break;
		end
	end
	if playerState.spellHasteBuffActive then
		playerState.buffedSpellHaste = playerState.spellHaste;
	else
		playerState.buffedSpellHaste = (1 + playerState.spellHaste)*1.05 - 1;
	end
	
	playerState.tidalWaves = tidalWavesUp;
	if ascendanceUp and not playerState.ascendanceAura then
		if playerState.ascendance then
			rss.ProcessCopyRecord(playerState.ascendance, fight);
		end
		playerState.ascendance = rss.CreateCopyRecord();
	end
	playerState.ascendanceAura = ascendanceUp;
	
	if guidanceUp and not playerState.guidanceAura then
		if playerState.ancestralGuidance then
			rss.ProcessCopyRecord(playerState.ancestralGuidance, fight);
		end
		playerState.ancestralGuidance = rss.CreateCopyRecord();
	end
	playerState.guidanceAura = guidanceUp;
	
	return playerState;
end

-- called when a spellcast is sent to server
-- this initiates a cast, since by definition no effects can happen before the spell is sent
-- we know the name, but not the spellID, which is only returned if the cast succeeds
function rss.SpellcastSent(...)
	if not rss.casts then return; end
	local spellName = select(2, ...);
	local targetName = select(4, ...);
	targetName = #targetName > 0 and targetName or nil; -- blank targets here are nil in combat log...
	local lineID = select(5, ...);
	local sentEventTime = GetTime();
	local haste = playerState.buffedSpellHaste;
	local targetHaste = { };
	for i, breakpoint in ipairs(rss.playerState.hasteBreakpoints) do
		targetHaste[i] = rss.CalcBuffedHasteFromRating(breakpoint.rating);
	end
	local unitHaste = rss.CalcBuffedHasteFromRating(rss.playerState.hasteRating+1) - rss.CalcBuffedHasteFromRating(rss.playerState.hasteRating);
	local cast = { sentEventTime = sentEventTime , spellName = spellName , haste = haste , targetHaste = targetHaste, unitHaste = unitHaste , lineID = lineID , targetName = targetName };
	rss.casts:Prepend(cast);
	rss.casts:Visit(rss.PruneCast, sentEventTime);
end

-- called when a spellcast succeeds
-- effects of the spellcast may have already begun to appear in the combat log before this event is received
-- 
function rss.SpellcastSucceeded(fight, ...)
	if not rss.casts then return; end
	local spellName = select(2, ...);
	local lineID = select(4, ...);
	local spellID = select(5, ...);
	local succeedEventTime = GetTime();
	rss.casts:Visit(rss.MatchCast, spellName, lineID, fight, spellID, succeedEventTime);
	rss.casts:Visit(rss.PruneCast, succeedEventTime);
end

-- Called to fill in success information for matching cast, or trash it if we don't care about the spell
function rss.MatchCast(list, node, spellName, lineID, fight, spellID, succeedEventTime)
	if node.spellName == spellName and node.lineID == lineID and not node.succeedEventTime then
		if rss.SpellFilter(spellID) then
			node.fight = fight;
			node.spellID = spellID;
			node.succeedEventTime = succeedEventTime;
			local updatePointer = math.floor((playerState.busyPointer + rss.WINDOW_FUTURE * rss.WINDOW_RATE) % rss.WINDOW_BINS);
			playerState.busyRing[updatePointer].casts:Append({ cast = node });
			rss.CorrelateEvents();
		else
			list:Remove(node);
			node = nil;
		end
		return true;
	end
end

-- runs the correlator function to test every event to a specified cast
-- if exclusive is true, the comparison stops after the first successful match
-- any matched events are deleted from the event list
function rss.CorrelateCastToEvents(events, event, cast, correlator, exclusive)
	if correlator(cast, event) then
		events:Remove(event);
		return exclusive;
	end
end

-- Called to remove casts that were sent but have not succeeded after some time, or casts that have had their busy fraction filled in
function rss.PruneCast(list, node, eventTime)
	if not node.succeedEventTime and eventTime > node.sentEventTime + rss.CAST_TIMEOUT then
		list:Remove(node);
		node = nil;
	elseif node.succeedEventTime and node.succeedLogTime and node.busyFraction then
		local finished = (eventTime > node.lastEventTime + rss.CAST_TIMEOUT) and (node.activeAuras == nil or (rss.QueueInterCombat and rss.QueueEndOfCombat) );
		if finished then
			rss.CastTotalHealing(node);
			--rss.DumpCast(node);
			rss.VaryTicks(node);
			rss.VaryCasts(node);
			for i, breakpoint in ipairs(rss.playerState.hasteBreakpoints) do
				if breakpoint.rating ~= rss.playerState.baseHasteRating then
					node.fight.incrementalHaste[i] = node.fight.incrementalHaste[i] +
						((node.deltaTicks and node.deltaTicks[i] or 0) + (node.deltaCasts and node.deltaCasts[i] or 0))/(breakpoint.rating - rss.playerState.baseHasteRating);
				end
			end
			node.fight.incrementalHaste[0] = node.fight.incrementalHaste[0] + (node.deltaCasts and node.deltaCasts[0] or 0);
			list:Remove(node);
			node = nil;
		end
	elseif (rss.QueueInterCombat and rss.QueueEndOfCombat) then
		list:Remove(node);
		node = nil;
	end
end

function rss.DumpCast(node)
	print(("%s: (%.4f/%.4f) actual/target Haste; (%d/%d) effective healing; %d copied"):format(node.spellName, node.haste, node.targetHaste[10], node.healing, node.rawHealing, node.originalHealing or 0));
	if node.events then
		local start = node.events:Start();
		while start do
			local spellName = start.event.spellId ~= 0 and select(1,GetSpellInfo(start.event.spellId)) or select(1,GetSpellInfo(node.spellID));
			print(start.event.eventType.." "..spellName.." -> ("..start.event.healed.."/"..start.event.amount..")");
			start = start.forward;
		end
	end
end

function rss.CastTotalHealing(node)
	local healing = 0.0;
	local rawHealing = 0.0;
	local copiedHealing = nil;
	if node.events then
		local start = node.events:Start();
		while start do
			-- Loop over events for this cast
			if start.event.eventType == rss.EVENTS.SPELL_HEAL or start.event.eventType == rss.EVENTS.SPELL_PERIODIC_HEAL then
				healing = healing + start.event.healed;
				rawHealing = rawHealing + start.event.amount;
				if start.event.copiedBy then
					if not copiedHealing then
						copiedHealing = rss.LinkedList:Create();
					end
					local copy = start.event.copiedBy:Start();
					while copy do
						-- Loop over spells copying this event
						local copyCast = copy.copier;
						local total = copiedHealing:Start();
						local copyTotal = nil; 
						while total do
							-- Loop over the spells we've already seen copying this *cast* (if any)
							if (total.cast == copyCast) then
								-- Saw this copier already
								copyTotal = total;
								break;
							end
							total = total.forward;
						end
						if not copyTotal then
							-- Have not seen this copier already
							copyTotal = {cast = copyCast , totalCopied = 0.0 };
							copiedHealing:Append(copyTotal);
						end
						copyTotal.totalCopied = copyTotal.totalCopied + start.event.amount;
						copy = copy.forward;
					end
				end
			end
			start = start.forward;
		end
	end
	node.healing = healing;
	node.rawHealing = rawHealing;
	node.copiedHealing = copiedHealing;
end

-- Busy ring maintenance
function rss.InitializeRing()
	local busyRing = { };
	for i = 0, rss.WINDOW_BINS-1 do
		busyRing[i] = { busy = 0 , casts = rss.LinkedList:Create() };
	end
	playerState.busyCounts = 0;
	playerState.busyPointer = 0;
	playerState.busyElapsed = 0.0;
	playerState.busyRing = busyRing;
end

function rss.SampleBusy()
	playerState.busy = (UnitCastingInfo("player") or GetSpellCooldown(rss.LIGHTNING_SHIELD) > 0 or UnitChannelInfo("player")) and 1 or 0;
	rss.UpdateRing(playerState.busy);
	if math.floor(playerState.busyPointer % rss.WINDOW_RATE) == 0 then 	-- keep data flowing at least once/second, during lulls/out-of-combat
		local globalTime = GetTime();
		if rss.metaGlobalTimestamp and globalTime > rss.metaGlobalTimestamp + 1 then
			rss.FlushMetaBuffer();
		end
		rss.casts:Visit(rss.PruneCast, globalTime); 
	end
end

function rss.UpdateRing(state)
	playerState.busyCounts = playerState.busyCounts - playerState.busyRing[playerState.busyPointer].busy;
	playerState.busyRing[playerState.busyPointer].busy = state;
	playerState.busyCounts = playerState.busyCounts + state;
	if playerState.busyRing[playerState.busyPointer].casts:Start() then
		playerState.busyRing[playerState.busyPointer].casts:Visit(rss.UpdateCastActivity, playerState.busyCounts/rss.WINDOW_BINS);
		rss.casts:Visit(rss.PruneCast, GetTime());
	end
	playerState.busyPointer = math.floor((playerState.busyPointer + 1) % rss.WINDOW_BINS);
end

function rss.UpdateCastActivity(list, node, busyFraction)
	node.cast.busyFraction = math.max(0.0, math.min(1.0, busyFraction));
	list:Remove(node);
	node = nil;
end

-- health table maintenance
function rss.ClearHealthTable()
	if UnitHealthMax("player") > 0 then
		healthTable = { [UnitGUID("player")] = { maxHealth = UnitHealthMax("player") , currentHealth = UnitHealth("player") } };
	else
		healthTable = { };
	end
	rss.healthTable = healthTable;
end
-- respond to HEALTH_CHANGED and MAX_HEALTH_CHANGED events for party/raid members
function rss.UpdateHealthTable(...)
	local unitID = select(1,...);
	if unitID == nil then return; end
	if UnitIsPlayer(unitID) or UnitPlayerOrPetInParty(unitID) or UnitPlayerOrPetInRaid(unitID) then
		rss.UpdateUnitHealth(unitID);
	end
end
-- general routine to add any unit we can heal to the health table
function rss.UpdateUnitHealth(unitID)
	if not UnitCanAssist("player", unitID) then return; end
	local unitGUID = UnitGUID(unitID);
	if unitGUID == nil then return; end
	local unitHealthTable = healthTable[unitGUID];
	if unitHealthTable == nil then
		unitHealthTable = { };
		healthTable[unitGUID] = unitHealthTable;
	end
	unitHealthTable.maxHealth = UnitHealthMax(unitID);
	unitHealthTable.currentHealth = UnitHealth(unitID);
	if unitHealthTable.maxHealth < 1 then
		healthTable[unitGUID] = nil;
	end
end

local spellSchools = { }; -- lookup table to cache expensive logarithm calculation 

-- look up the critical chance for a particular spell/school
function rss.GetCriticalChance(spellId, spellSchool)
	local schoolIndex;
	-- check for override (Flames of Life)
	if rss.SPELLS[spellId] and rss.SPELLS[spellId].school then
		schoolindex = rss.SPELLS[spellId].school;
	else
		if not spellSchools[spellSchool] then
			spellSchools[spellSchool] =  math.floor(1+log(spellSchool)/log(2));
		end
		schoolIndex = spellSchools[spellSchool];
	end
	local myCriticalChance = rss.playerState.criticalChance[schoolIndex] or rss.playerState.criticalChance[rss.SCHOOL_NATURE];
	if spellId == rss.HEALING_SURGE then                                -- is this a Healing Surge critical?
		if rss.playerState.tidalWaves then              					-- tested: even if this cast consumed the last TW stack, we still have the buff at this moment
			myCriticalChance = math.min(1.0, myCriticalChance + 0.3);   -- adjust HS critical chance for TW
		end
	end
	return myCriticalChance;
end

-- check for a crit-enhancing metagem
function rss.CheckMeta()
	local gem1, gem2, gem3 = GetInventoryItemGems(GetInventorySlotInfo("HeadSlot")); -- head slot
	if tContains(rss.CRITICAL_BONUS_GEMS, gem1) or 
	   tContains(rss.CRITICAL_BONUS_GEMS, gem2) or 
	   tContains(rss.CRITICAL_BONUS_GEMS, gem3) then
		playerState.criticalMultiplier = rss.CRITICAL_MULTIPLIER * 1.03;
	else
		playerState.criticalMultiplier = rss.CRITICAL_MULTIPLIER;
	end
end

-- Find the item level of something equipped in a slot (including upgrades, which is why this isn't as easy as it should be)
function rss.GetItemLevel(slot)
	local g=MyTooltipScan or CreateFrame("GameTooltip","MyTooltipScan",UIParent,"GameTooltipTemplate")
	g:SetOwner(UIParent,"ANCHOR_NONE")
	g:SetInventoryItem("player",slot)
	for i=1,g:NumLines() do
		local line = _G["MyTooltipScanTextLeft"..i]:GetText();
		local ilvl = line:match('%d%d%d$');
		if ilvl then
			return tonumber(ilvl);
		end
	end
end

-- Scan equipped gear for tier 15 pieces and the Lei-Shen trinket
function rss.CheckGear()
	rss.CheckMeta();

	playerState.aaSetBonusT15 = false;
	local n15 = 0;
	for slotName, tierItems in pairs(rss.SET_BONUS_T15) do
		local equipID = GetInventoryItemID("player", GetInventorySlotInfo(slotName));
		if equipID and tContains(tierItems, equipID) then
			n15 = n15 + 1;
		end
	end
	if n15 >= 4 then
		playerState.aaSetBonusT15 = true;
	end
	
	local chaliceLevel = nil;
	for slotName, trinkets in pairs(rss.LIGHTNING_IMBUED_CHALICE) do
		local equipID = GetInventoryItemID("player", GetInventorySlotInfo(slotName));
		if equipID and tContains(trinkets, equipID) then
			chaliceLevel = rss.GetItemLevel(GetInventorySlotInfo(slotName));
			break;
		end
	end
	-- fill the appropriate base heal determined by the item level
	if chaliceLevel and rss.CHALICE_BASE_HEAL[chaliceLevel] then
		local baseHeal = rss.CHALICE_BASE_HEAL[chaliceLevel];
		local coeff = rss.SPELLS[rss.PROC_RESTORATION_INFINITE]["SPELL_HEAL"];
		coeff[1] = baseHeal;
		coeff[2] = baseHeal;
	end
end

function rss.CheckTalents()
	local oldTalents = playerState.ancestralSwiftnessActive;
	if select(5, GetTalentInfo(rss.TALENT_ANCESTRAL_SWIFTNESS)) then
		playerState.ancestralSwiftnessActive = true;
	else
		playerState.ancestralSwiftnessActive = false;
	end
	if playerState.ancestralSwiftnessActive ~= oldTalents or not playerState.hasteBreakpoints then
		playerState.hasteBreakpoints = rss.CalcHasteBreakpoints();
	end
end